/*
 * Author:	Brian J. Murrell <brian@interlinx.bc.ca>
 *
 * Stateful Packet Filter -- constantly adjust an inbound packet filter
 * to deal with allowing traffic in based on what was sent out.
 *
 * RealAudio support added by Marc Boucher <marc@CAM.ORG>
 * (based on ip_masq_raudio.c by Nigel Metheringham)
 *
 *	This program is free software; you can redistribute it and/or modify
 *	it under the terms of the GNU General Public License as published by
 *	the Free Software Foundation; either version 2 of the License, or
 *	(at your option) any later version.
 *
 *	This program is distributed in the hope that it will be useful,
 *	but WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *	GNU General Public License for more details.
 *
 *	You should have received a copy of the GNU General Public License
 *	along with this program; if not, write to the Free Software
 *	Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

#ifndef SPF_VIANAME
#define SPF_VIANAME "" /* or for example "eth1" */
#endif

#include <stdio.h>
#include <errno.h>
#include <string.h>
#include <stdarg.h>
#include <syslog.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <signal.h>
#include <fcntl.h>
#include <unistd.h>
#include <netinet/in.h>
#include <netinet/ip.h>
#include <netinet/ip_icmp.h>
/*
#include <netinet/ip_tcp.h>
*/
#include <arpa/inet.h>
#include <limits.h>
#include "kernel_headers.h"

struct fwhdr {
	__u32 len;
	__u32 mark;
	char interface[IFNAMSIZ];
};

#ifndef MAX
#define MAX(_a,_b)	((_a)<(_b)?(_b):(_a))
#endif
#ifndef MIN
#define MIN(_a,_b)	((_a)<(_b)?(_a):(_b))
#endif

/* this should be largest value a time_t can hold without overflowing */
#define MAX_TIMEOUT	(LONG_MAX/CLK_TCK)
#define LONG_TIMEOUT	0x7FFFFFFFL;
#define	DEFAULT_ACTION	"DENY"	
#define TRACEROUTE_MIN	33435
#define TRACEROUTE_MAX	TRACEROUTE_MIN+30*3
/* what rule to start inserting dyn. rules */
#if defined(ALLOW_EST_TCP) && !defined(ALLOW_EST_TCP_NORULE)
#define START_DYN_RULES	2	
#else
#define START_DYN_RULES	1
#endif

static time_t next_timeout;	/* absolute time of next select loop process */

struct connection {
	struct in_addr src_addr;
	struct in_addr src_mask;
	struct in_addr dst_addr;
	struct in_addr dst_mask;
	uint proto;
	uint src_port_lo;
	uint src_port_hi;
	uint dst_port_lo;
	uint dst_port_hi;
	char interface[IFNAMSIZ];
	time_t timeout;
	struct connection *prev;
	struct connection *next;
	void *data;		/* store connection data across packets */
};

struct connection *connections=NULL, *last_connection, *first_connection;

static uint debuglev=0;
static uint asdaemon=1;
static uint timer_stat=0;

/* timers */
#define CLEAR_TIMER(timer) timer.tv_sec=0; timer.tv_usec=0;
struct timeval timer_select, timer_setup_select, timer_expire,
	       timer_read_fwdata, timer_do_packet, timer_misc;

static struct in_addr hostmask = { 0xFFFFFFFF };
static struct in_addr zeroaddr = { 0 };
/* bjm - we should determine the address for our subnet automagically */
/* don't forget about yuccky endianess! */
static struct in_addr subnetaddr = { 0x000035D1 };
static struct in_addr subnetmask = { 0x00C0FFFF };

#define add_tcp_pr(sa, da, spl, sph, dpl, dph, s, i) add_rule(sa, hostmask, da, hostmask, spl, sph, dpl, dph, IPPROTO_TCP, s, i)
#define add_udp_pr(sa, da, spl, sph, dpl, dph, s, i) add_rule(sa, hostmask, da, hostmask, spl, sph, dpl, dph, IPPROTO_UDP, s, i)
#define add_icmp_r(sa, da, tl, th, cl, ch, s, i) add_rule(sa, hostmask, da, hostmask, tl, th, cl, ch, IPPROTO_ICMP, s, i)
#define add_tcp_mask(sa, sm, da, dm, sp, dp, s, i) add_rule(sa, sm, da, dm, sp, sp, dp, dp, IPPROTO_TCP, s, i)
#define add_udp_mask(sa, sm, da, dm, sp, dp, s, i) add_rule(sa, sm, da, dm, sp, sp, dp, dp, IPPROTO_UDP, s, i)
#define add_icmp_mask(sa, sm, da, dm, t, c, s, i) add_rule(sa, sm, da, dm, t, t, c, c, IPPROTO_ICMP, s, i)
#define add_tcp(sa, da, sp, dp, s, i) add_rule(sa, hostmask, da, hostmask, sp, sp, dp, dp, IPPROTO_TCP, s, i)
#define add_udp(sa, da, sp, dp, s, i) add_rule(sa, hostmask, da, hostmask, sp, sp, dp, dp, IPPROTO_UDP, s, i)
#define add_icmp(sa, da, t, c, s, i) add_rule(sa, hostmask, da, hostmask, t, t, c, c, IPPROTO_ICMP, s, i)

extern uint nsaddrs, ndaddrs;
extern struct in_addr *saddrs, *daddrs;
extern int flush_entries(const ip_chainlabel chain);
void hdump(register const u_char *, register u_int);
int debug(char *, ...);
struct connection *find_connection(struct in_addr, struct in_addr,
		struct in_addr, struct in_addr, uint, uint, uint, uint, uint,
		char *);
struct connection *record_connection(struct in_addr, struct in_addr,
		struct in_addr, struct in_addr, uint, uint, uint, uint,
		uint, time_t, char *);
void add_rule(struct in_addr src, struct in_addr src_mask, struct in_addr dst,
	 struct in_addr dst_mask, int src_port_lo, int src_port_hi,
	 int dst_port_lo, int dst_port_hi, uint proto, time_t secs,
	 char *interface);
void cleanup_exit(int);
void catch_signal(int);
void cleanup_chains();
void set_timer(struct timeval *timer);
char *timer2str(struct timeval *timer);

int
main(int argc, char *argv[]) {

	char packet[65555];
	int stream, c;
	ssize_t num_read;
	long sec, usec;

	openlog("spf", LOG_CONS|LOG_NDELAY|LOG_PID, LOG_AUTHPRIV);
	while ((c=getopt(argc, argv, "sx:"))!=EOF)
		switch (c) {
			case 'x':
				debuglev=atoi(optarg);
				asdaemon=0;
				setlinebuf(stdout);
				setlinebuf(stderr);
				break;
			case 's':
				asdaemon=0;
				break;
			default:
				cleanup_exit(SIGTERM);
		}
	if ((stream=open("/dev/fwdata", O_RDONLY))<0) {
		perror("open /dev/fwdata");
		exit(1);
	}

	/* dump timer stats on a SIGUSR1 */
	signal(SIGUSR1, catch_signal);

	/* cleanup and exit when we get various signals */
	signal(SIGINT, catch_signal);
	signal(SIGTERM, catch_signal);
	signal(SIGSEGV, catch_signal);
	signal(SIGPIPE, catch_signal);	/* in case debug is being piped */

	/* install filters needed to do stateful filtering */
	install_spf_filters();

	if (asdaemon)
		/* now fork and exit */
		if ((c=fork())<0) {	/* fork failed */
			perror("fork");
			exit(1);
		} else if (c==0) 	/* child */
			setsid();
		else 			/* parent */
			exit(0);

	next_timeout=LONG_TIMEOUT;
	/* clear the timers */
	CLEAR_TIMER(timer_select);
	CLEAR_TIMER(timer_setup_select);
	CLEAR_TIMER(timer_expire);
	CLEAR_TIMER(timer_read_fwdata);
	CLEAR_TIMER(timer_do_packet);
	CLEAR_TIMER(timer_misc);
	set_timer(NULL);	/* reset the timer */
	while(1) {
		int ret;
		fd_set readfds;
		struct timeval timeout;

		FD_ZERO(&readfds);
		FD_SET(stream, &readfds);
		timeout.tv_usec=0;
		timeout.tv_sec=next_timeout-time((time_t)NULL);
		/* NOTE: Short-circuit the timeout mode below in case our
		 *       read source is so busy that select never times out.
		 *       Could it really happen?  I think so.
		 */
		if (timeout.tv_sec<1) {
			set_timer(&timer_setup_select);
			/* check the connections table for expired entries */
			expire_connections();
			set_timer(&timer_expire);
		}
		/* kernels > 2.1.126 limit the schedule_timeout() parameter
		 * (in ticks) to a signed long.
		 */
		timeout.tv_sec=MIN(MAX_TIMEOUT, timeout.tv_sec);
		set_timer(&timer_setup_select);
		if ((ret=select(stream+1, &readfds, NULL, NULL, &timeout))<0) {
			set_timer(&timer_select);
			if (errno != EINTR) {
				perror("select");
				cleanup_exit(1);
			}
		} else if (ret==0) {
/* is there a possibility that we could be so busy to never get here?  If
 * there were (say) _always_ data on the netlink device, such that select
 * always returned > 0 we could in theory miss out on expiring connections.
 * I have never seen this happen mind you.
 * What does select do if you call it with a timeout of 0, yet there are fds
 * that need processing?  Well a test reveals that it returns > 0.  Darn!
 */
			set_timer(&timer_select);
			/* check the connections table for expired entries */
			expire_connections();
			set_timer(&timer_expire);
		} else if (ret>0) {
			set_timer(&timer_select);
			if (FD_ISSET(stream, &readfds)) {
				if ((num_read=read(stream, packet,
					      sizeof(char)*2000))<0) {
					set_timer(&timer_read_fwdata);
					if (errno == ENOBUFS) {
						perror("netlink queue overflow "
						       "(pressing on)");
						continue;
					} else {
						perror("read");
						cleanup_exit(1);
					}
				}
				set_timer(&timer_read_fwdata);
				do_packet(packet, num_read);
				set_timer(&timer_do_packet);
			}
		}
		debug("\n");
		if (debuglev>3)
			display_connections();
		set_timer(&timer_misc);

		if (timer_stat) {
			timer_stat=0;
			sec=timer_select.tv_sec+timer_setup_select.tv_sec+
			    timer_expire.tv_sec+timer_read_fwdata.tv_sec+
			    timer_do_packet.tv_sec+timer_misc.tv_sec;
			usec=timer_select.tv_usec+timer_setup_select.tv_usec+
			     timer_expire.tv_usec+timer_read_fwdata.tv_usec+
			     timer_do_packet.tv_usec+timer_misc.tv_usec;
			sec+=usec/1000000;
			usec+=usec%1000000;
			fprintf(stderr, "  select    setup     expire     "
				"read   do_packet   misc        TOTAL\n");
			fprintf(stderr, "  (idle)    select    conns     "
				"fwdata             timers\n");
			fprintf(stderr, "%s ", timer2str(&timer_select));
			fprintf(stderr, "%s ", timer2str(&timer_setup_select));
			fprintf(stderr, "%s ", timer2str(&timer_expire));
			fprintf(stderr, "%s ", timer2str(&timer_read_fwdata));
			fprintf(stderr, "%s ", timer2str(&timer_do_packet));
			fprintf(stderr, "%s ", timer2str(&timer_misc));
			fprintf(stderr, "%d.%07ld\n", sec, usec);
			set_timer(&timer_misc);
		}
	}
}

void
cleanup_chains() {

	for_each_chain(flush_entries);
	/* remove our statinpt chain */
	if (delete_chain("statinpt", 0)<0)
		perror("removing statinpt chain");

}

void
catch_signal(int sig) {

	if (sig==SIGUSR1) {
		debug("Got signal to dump stats\n", sig);
		timer_stat=1;
		return;
	}
	debug("Got signal %d, cleaning up and exiting\n", sig);
	cleanup_chains();
	if (sig==SIGSEGV)
		abort();	/* leave a core */
	closelog();
	exit(0);

}

/* Flush all rules in all chains and exit */
void
cleanup_exit(int status) {

	debug("flushing chains and exiting\n");
	cleanup_chains();
	closelog();
	exit(status);
}

#define BYTESTR 16
#define BYTELEN  4

void
hdump(register const u_char *bp, register u_int length)
{

/* this could probably be more elegant, but it was ripped off from "hdump"
 * without much "how does it work" questions asked
 */
	long pos;
	size_t nread;
	register const u_char *sp;
	register u_int i;
	int flen = BYTELEN;
	register int s;
	struct timeval tv;
	struct timezone tz;

	gettimeofday(&tv, &tz);
	s=(tv.tv_sec-8*60*60)%86400;
	debug("\n%02d:%02d:%02d.%06u\n", s/3600, (s%3600)/60, s%60, tv.tv_usec);
	if (length<0) {
		/* can't print if the data size is less that the offset */
		return;
	}
#if 0
	if ((long)bp & 1) {
		default_print_unaligned(bp, length);
		return;
	}
#endif
	pos = 0;
	while ( length > 0 ) {
		if (length < BYTESTR)
			nread = length;
		else
			nread = BYTESTR;
		sp = bp + pos;

		printf ( "\n%8ld:  ", pos );

		for ( i = 0; i < BYTESTR; i++ ) {
			if ( i < nread )
				printf ( "%02x", sp[i] );
			else
				printf ( "  " );
			if ( ( i % flen ) == ( flen - 1 ) )
				printf ( " " ); 
		}

		printf ( "  " ); 
		for ( i = 0; i < BYTESTR; i++ ) {
			if ( i < nread ) {
				if ( isprint ( sp[i] ) )
					printf ( "%c", sp[i] );
				else
					printf ( "." );
			}
		}

		pos += nread;
		length -= nread;

	}
	printf("\n");
}

int
debug(char *format, ...) {

	if (debuglev>0) {
		va_list ap;

		va_start(ap, format);
		(void)vfprintf(stderr, format, ap);
		va_end(ap);
	}

}

do_packet(const char *p, int num_read) {

	char *ip_src, *ip_dst;
	struct ip *ip;

	(char *)ip=p+sizeof(struct fwhdr);

	if (debuglev>0) {
		ip_src=strdup(inet_ntoa(ip->ip_src));
		ip_dst=strdup(inet_ntoa(ip->ip_dst));
		debug("sent: (%d) %s > %s (frag:%lu)", num_read, ip_src, ip_dst, ip->ip_off);
		if (debuglev>2)
			hdump(p, num_read);
		free(ip_src);
		free(ip_dst);
	}

	/* ignore IP fragments */
	if ((ntohs(ip->ip_off)&IP_OFFMASK)>0) 
		return(0);
	switch(ip->ip_p) {
		case IPPROTO_TCP:
			do_tcp(p, num_read - sizeof(struct fwhdr));
			break;
		case IPPROTO_UDP:
			do_udp(p);
			break;
		case IPPROTO_ICMP:
			do_icmp(p);
			break;
		default:
			fprintf(stderr, "unknown IP protocol \"%d\"\n",
				ip->ip_p);
	}

}

int
do_tcp(char *p, int len) {

	struct ip *ip;
	struct tcphdr *tcp;
	struct fwhdr *fw=(struct fwhdr *)p;
	char *data, *data_limit;

	debug(" tcp");

	(char *)ip=p+sizeof(struct fwhdr);
	(char *)tcp=(char *)ip+ip->ip_hl*4;
	debug(" %d %d", ntohs(tcp->source), ntohs(tcp->dest));

#ifndef ALLOW_EST_TCP
	/*
	 * Allow return TCP packets
	 */
	/* install a reverse rule for 30 seconds */
	add_tcp(ip->ip_dst, ip->ip_src, ntohs(tcp->dest),
		ntohs(tcp->source), 30, fw->interface);
#endif
	
	data=(char *)tcp+tcp->doff*4;
	data_limit = p + len;
	/* Watch for and open holes for FTP DATA sessions */
	if (ntohs(tcp->dest)==21) {
		if (debuglev>2)
			hdump(data,
			      ntohs(ip->ip_len)-(ip->ip_hl*4)-(tcp->doff*4));
		/* this is purely *cheating* cheapo code.  assuming this stuff
		 * will always be at byte 0 of a data packet is naive.  the
		 * proper solution is to keep a running buffer which we always
		 * search for what we want
		 */
		if (strncasecmp(data, "port ", 4)==0) {
			/* parse the port specification */

			unsigned short int port;
			char *c, *p;
			int l;

			l = ntohs(ip->ip_len)-(ip->ip_hl*4)-(tcp->doff*4) - 1;
			if(l >= len)
				l = len - 1;
			data[l] = '\0'; /* strrchr needs null-terminated buf */
			p=(char *)&port;
			c=strrchr(data, ',');
			*c++='\0';
			p[0]=strtol(c, (char **)NULL, 10);
			c=strrchr(data, ',');
			*c++='\0';
			p[1]=strtol(c, (char **)NULL, 10);
			add_tcp(ip->ip_dst, ip->ip_src, 20, port, 30,
				fw->interface);
		}
	} else if (ntohs(tcp->source)==21) {
		if (strncasecmp(data, "227 Entering Passive Mode", 25)==0) {
			/* parse the port specification */

#if 1 /* old - bad: we don't count on the response being split across packets.
       *            I am not sure of the best way to deal with this yet.
       *            See below for an idea.
       */
			unsigned short int port;
			char *c, *p=(char *)&port;
			int l;

			l = ntohs(ip->ip_len)-(ip->ip_hl*4)-(tcp->doff*4) - 1;
			if(l >= len)
				l = len - 1;
			data[l] = '\0'; /* strrchr needs null-terminated buf */
			if ((c=strrchr(data, ')'))==NULL) {
				syslog(LOG_NOTICE, "Invalid PASV port "
						   "specification \"%s\"",
				       data);
				return;
			}
			*c='\0';
			if ((c=strrchr(data, ','))==NULL) {
				syslog(LOG_NOTICE, "Error processing PASV port "
						   "specification \"%s\" at "
						   "%s:%d", __FILE__, __LINE__,
				       data);
				return;
			}
			*c++='\0';
			p[0]=strtol(c, (char **)NULL, 10);
			if ((c=strrchr(data, ','))==NULL) {
				syslog(LOG_NOTICE, "Error processing PASV port "
						   "specification \"%s\" at "
						   "%s:%d", __FILE__, __LINE__,
				       data);
				return;
			}
			*c++='\0';
			p[1]=strtol(c, (char **)NULL, 10);
			add_tcp_pr(ip->ip_dst, ip->ip_src, 0, 65535, port,
				   port, 30, fw->interface);
#else /* new - good, not finished
       *            I think if we create a "connection" record, and keep adding
       *            received data until we have received a full line (term-
       *            inated with the \r\n) and then parse it we will be OK.
       */
			struct connection *conn;
			int l;

			l = ntohs(ip->ip_len)-(ip->ip_hl*4)-(tcp->doff*4) - 1;
			if(l >= len)
				l = len - 1;
			data[l] = '\0'; /* strrchr needs null-terminated buf */
			if (data[l-1]=='\r' && data[l-2]=='\r') {
				/* get the host and port and open a rule
				return;
			}
			conn=record_connection(ip->ip_src, hostmask,
					       ip->ip_dst, hostmask, ip->ip_p,
				  	       ntohs(tcp->source),
					       ntohs(tcp->source),
					       ntohs(tcp->dest),
					       ntohs(tcp->dest),
				  	       -1, fw->interface);	
			/* store what we have seen so far */
			(char *)conn->data=strdup(data);
		} else {
			/* see if this is another portion of the above */
 			if ((conn=find_connection(ip->ip_src, hostmask,
						  ip->ip_dst, hostmask,
			     			  ip->ip_p,
						  ntohs(tcp->source),
			     			  ntohs(tcp->source),
						  ntohs(tcp->dest),
			     			  ntohs(tcp->dest),
						  fw->interface))==NULL)
				return;
			/* found one.  see if it's complete and process it */
			/* don't forget to free the connection and it's data */
#endif
		}
	} else if ((ntohs(tcp->dest)==554) || (ntohs(tcp->dest)==7070)) {
		unsigned short version, msg_id, msg_len, udp_port;

		if(!memcmp(data, "OPTIONS", 7) || !memcmp(data, "DESCRIBE", 8)) {
			/* This is an RTSP client (not supported for now) */
			return 0;
		}
		
		if(memcmp(data, "PNA", 3)) {
			return 0; /* not initial protocol packet - ignored */
		}

		data += 3;

		version = ntohs(*(unsigned short *)data);
		debug("RealAudio: initial seen - protocol version %d\n", version);
		if(version >= 256) {
			debug("RealAudio: protocol version %d not supported\n", version);
		}

		data += 2;

		while (data+4 < data_limit) {
			memcpy(&msg_id, data, 2);
			data += 2;
			memcpy(&msg_len, data, 2);
			data += 2;
			if (ntohs(msg_id) == 0) {
				/* The zero tag indicates the end of options */
				debug("RealAudio: packet end tag seen\n");
				return 0;
			}
			debug("RealAudio: msg %d - %d byte\n",
			       ntohs(msg_id), ntohs(msg_len));
			if (ntohs(msg_id) == 0) {
				/* The zero tag indicates the end of options */
				return 0;
			}
			p = data;
			data += ntohs(msg_len);
			if (data > data_limit)
			{
				debug("RealAudio: Packet too short for data\n");
				return 0;
			}
			if ((ntohs(msg_id) == 1) || (ntohs(msg_id) == 7)) {
				/* 
				 * MsgId == 1
				 * Audio UDP data port on client
				 *
				 * MsgId == 7
				 * Robust UDP error correction port number on client
				 *
				 * Since these messages are treated just the same, they
				 * are bundled together here....
				 */
				memcpy(&udp_port, p, 2);
	
				/* 
				 * Sometimes a server sends a message 7 with a zero UDP port
				 * Rather than do anything with this, just ignore it!
				 */
				if (udp_port == 0)
					continue;
	
	
				/* install a reverse rule for 30 seconds */
				add_udp_pr(ip->ip_dst, ip->ip_src, 0, 65535, ntohs(udp_port), ntohs(udp_port), 30, fw->interface);
			}
		}
	}
}

int
do_udp(const char *p) {

	struct ip *ip;
	struct udphdr *udp;
	struct fwhdr *fw=(struct fwhdr *)p;
	unsigned int timeout;
	struct connection *conn;

	debug(" udp");

	timeout=30;	/* default timeout for UDP is 30 seconds */
	(char *)ip=p+sizeof(struct fwhdr);
	(char *)udp=(char *)ip+ip->ip_hl*4;

	/* outbound traceroutes should allow corresponding inbound "ttl timed
	 * out" messages to the tracerouting host for 15 seconds; as well,
	 * allow ICMP port unreachable from the tracrouted host for 15 secs
	 */
	if (ntohs(udp->dest) >= TRACEROUTE_MIN &&
	    ntohs(udp->dest) <= TRACEROUTE_MAX) {

 		add_icmp_mask(zeroaddr, zeroaddr, ip->ip_src, hostmask, 11, 0,
			      15, fw->interface);
		add_icmp(ip->ip_dst, ip->ip_src, 3, 3, 15, fw->interface);
		return(0);	/* don't need to allow the UDPs back in */
	} else if (ntohs(udp->dest)==53) {
		timeout=90;	/* timeout for DNS is 90 seconds */
	} else if (ntohs(udp->dest)==4000) {
		timeout=90;	/* timeout for ICQ is 60 seconds */
	} else if ((ntohs(udp->source)==68) && (ntohs(udp->dest)==67)) {
		/* DHCP request */
		/* we don't know where the response is going to come from
		 * although I suppose we could limit to our broadcast domain
		 * (our subnet)
		 */
		timeout=10;	/* timeout for DHCP is 10 seconds! */
 		add_udp_mask(subnetaddr, subnetmask, ip->ip_src, hostmask,
			     ntohs(udp->dest), ntohs(udp->source), 10,
			     fw->interface);
 		add_udp_mask(subnetaddr, subnetmask, hostmask, hostmask,
			     ntohs(udp->dest), ntohs(udp->source), 10,
			     fw->interface);
		return(0);
	}
	/*
	 * Allow return UDP packets
	 */
	
	debug(" %d %d", ntohs(udp->source), ntohs(udp->dest));
#if 0 /* not sure what the "new" code does with respect to "conn" */
	/* install a reverse rule for timeout seconds */
	add_udp(ip->ip_dst, ip->ip_src, ntohs(udp->dest), ntohs(udp->source),
		timeout, fw->interface);
#endif
 	/* check for previously allowed incoming connection first */
 	conn=find_connection(ip->ip_src, hostmask, ip->ip_dst, hostmask,
			     IPPROTO_UDP, ntohs(udp->source),
			     ntohs(udp->source), ntohs(udp->dest),
			     ntohs(udp->dest), NULL);
 	if(conn==NULL) {
 		conn=find_connection(ip->ip_src, hostmask, ip->ip_dst, hostmask,
				     IPPROTO_UDP, 0, 65535, ntohs(udp->dest),
				     ntohs(udp->dest), NULL);
 	}
 	if(conn) {
 		/* install a new rule for timeout seconds */
 		add_udp(ip->ip_src, ip->ip_dst, ntohs(udp->source),
			ntohs(udp->dest), timeout, conn->interface);
 	} else {
 		/* install a reverse rule for timeout seconds */
 		add_udp(ip->ip_dst, ip->ip_src, ntohs(udp->dest),
			ntohs(udp->source), timeout, fw->interface);
 	}

}

int
do_icmp(char *p) {

	struct ip *ip;
	struct icmphdr *icmp;
	struct fwhdr *fw=(struct fwhdr *)p;
	struct connection *conn;

	debug(" icmp");

	(char *)ip=p+sizeof(struct fwhdr);
	(char *)icmp=(char *)ip+ip->ip_hl*4;
	/*
	 * Allow ICMP replies matching requests
	 */
	
	if (icmp->type==8 && icmp->code==0) {
		/* install a reverse rule for 30 seconds */

		debug(" echo request");
		add_icmp(ip->ip_dst, ip->ip_src, 0, 0, 30, fw->interface);
	} else
		debug(" (%d, %d)", icmp->type, icmp->code);

}

/* allow a rule for a packet from a src to a dest for a time limit */
void
add_rule(struct in_addr src, struct in_addr src_mask, struct in_addr dst,
	 struct in_addr dst_mask, int src_port_lo, int src_port_hi,
	 int dst_port_lo, int dst_port_hi, uint proto, time_t secs,
	 char *interface) {

	char *ip_src, *ip_dst, *ip_src_mask, *ip_dst_mask;
	struct connection *conn;
	static struct ip_fwuser fw;
	int ret;

	/* update an existing entry if there is one */
	conn=NULL;
	if ((conn=find_connection(src, src_mask, dst, dst_mask, proto,
				  src_port_lo, src_port_hi, dst_port_lo,
				  dst_port_hi, interface))!=NULL) {
		conn->timeout=secs+time((time_t)NULL);
		return;
	}

	if (debuglev>0) {
		ip_src=strdup(inet_ntoa(src));
		ip_dst=strdup(inet_ntoa(dst));
		ip_src_mask=strdup(inet_ntoa(src_mask));
		ip_dst_mask=strdup(inet_ntoa(dst_mask));
		debug("add rule %s %s %u->%u %u->%u %u %s %lu", ip_src,
		       ip_dst, src_port_lo,
		       src_port_hi, dst_port_lo, dst_port_hi, proto,
		       interface, secs);
		/* don't leak memory! */
		free(ip_src);
		free(ip_src_mask);
		free(ip_dst);
		free(ip_dst_mask);
	}
	fw.ipfw.fw_src=src;
	fw.ipfw.fw_smsk=src_mask;
	fw.ipfw.fw_dst=dst;
	fw.ipfw.fw_dmsk=dst_mask;
	fw.ipfw.fw_proto=proto;
	fw.ipfw.fw_flg=0;
	fw.ipfw.fw_tosand = 0xFF;
	fw.ipfw.fw_tosxor = 0x00;
	fw.ipfw.fw_invflg=0;
	fw.ipfw.fw_spts[0]=src_port_lo;
	fw.ipfw.fw_spts[1]=src_port_hi;
	fw.ipfw.fw_dpts[0]=dst_port_lo;
	fw.ipfw.fw_dpts[1]=dst_port_hi;
	fw.ipfw.fw_redirpt=0;
	fw.ipfw.fw_outputsize=0;
	strcpy(fw.ipfw.fw_vianame, interface);
	strcpy(fw.label, "ACCEPT");
	nsaddrs=1;
	saddrs=&src;
	ndaddrs=1;
	daddrs=&dst;
	if ((ret=insert_entry("input", &fw, START_DYN_RULES, 0, 0))!=0) {
		fprintf(stderr, "error inserting a new input rule (ret=%d) at %s:%d\n0x%x:%u->%u -> 0x%x:%u->%u, proto %u\n", ret, __FILE__, __LINE__, src, src_port_lo, src_port_hi, dst, dst_port_lo, dst_port_hi, proto);
	} else {
		record_connection(src, src_mask, dst, dst_mask, proto,
				  src_port_lo, src_port_hi, dst_port_lo,
				  dst_port_hi, secs, interface);	
	}

}

/* remove a rule */
delete_rule(struct connection *conn) {

	static struct ip_fwuser fw;
	int ret;

	memcpy(&fw.ipfw.fw_src, &conn->src_addr, sizeof(fw.ipfw.fw_src));
	memcpy(&fw.ipfw.fw_smsk, &conn->src_mask, sizeof(fw.ipfw.fw_smsk));
	memcpy(&fw.ipfw.fw_dst, &conn->dst_addr, sizeof(fw.ipfw.fw_dst));
	memcpy(&fw.ipfw.fw_dmsk, &conn->dst_mask, sizeof(fw.ipfw.fw_dmsk));
	fw.ipfw.fw_proto=conn->proto;
	fw.ipfw.fw_flg=0;
	fw.ipfw.fw_tosand = 0xFF;
	fw.ipfw.fw_tosxor = 0x00;
	fw.ipfw.fw_invflg=0;
	fw.ipfw.fw_spts[0]=conn->src_port_lo;
	fw.ipfw.fw_spts[1]=conn->src_port_hi;
	fw.ipfw.fw_dpts[0]=conn->dst_port_lo;
	fw.ipfw.fw_dpts[1]=conn->dst_port_hi;
	fw.ipfw.fw_redirpt=0;
	fw.ipfw.fw_outputsize=0;
	strcpy(fw.ipfw.fw_vianame, conn->interface);
	strcpy(fw.label, "ACCEPT");
	nsaddrs=1;
	saddrs=&conn->src_addr;
	ndaddrs=1;
	daddrs=&conn->dst_addr;
	if ((ret=delete_entry("input", &fw, 0, 0))!=0) {
		fprintf(stderr, "error deleteing an input rule %s:%d\n",
			__FILE__, __LINE__);
	} else {
		delete_connection(conn);	
	}

}

struct connection *
record_connection(struct in_addr src, struct in_addr src_mask,
		  struct in_addr dst, struct in_addr dst_mask,
		  uint proto, uint src_port_lo, uint src_port_hi,
		  uint dst_port_lo, uint dst_port_hi, time_t timeout,
		  char *interface) {

	struct connection *new_conn;

	if (debuglev>1) {

		char *ip_src, *ip_dst, *ip_src_mask, *ip_dst_mask;

		ip_src=strdup(inet_ntoa(src));
		ip_dst=strdup(inet_ntoa(dst));
		ip_src_mask=strdup(inet_ntoa(src_mask));
		ip_dst_mask=strdup(inet_ntoa(dst_mask));
		debug("recording connection %s:%s %s:%s %u->%u %u->%u %u %lu\n",
		      ip_src, ip_src_mask, ip_dst, ip_dst_mask,
		      src_port_lo, src_port_hi, dst_port_lo,
		      dst_port_hi, proto, timeout);
		/* don't leak memory! */
		free(ip_src);
		free(ip_src_mask);
		free(ip_dst);
		free(ip_dst_mask);
	}
	/* make timeout absolute */
	timeout+=time((time_t)NULL);

	if ((new_conn=(struct connection *)malloc(sizeof(struct connection)))==
	    NULL) {
		perror("Allocating new connection");
		exit(1);
	}
	if (!first_connection)
		first_connection=new_conn;

	memcpy(&new_conn->src_addr, &src, sizeof(src));
	memcpy(&new_conn->src_mask, &src_mask, sizeof(src_mask));
	memcpy(&new_conn->dst_addr, &dst, sizeof(dst));
	memcpy(&new_conn->dst_mask, &dst_mask, sizeof(dst_mask));
	new_conn->proto=proto;
	new_conn->src_port_lo=src_port_lo;
	new_conn->src_port_hi=src_port_hi;
	new_conn->dst_port_lo=dst_port_lo;
	new_conn->dst_port_hi=dst_port_hi;
	new_conn->timeout=timeout;
	new_conn->next=NULL;
	new_conn->prev=last_connection;
	strcpy(new_conn->interface, interface);
	if (last_connection)
		last_connection->next=new_conn;
	last_connection=new_conn;
	next_timeout=MIN(next_timeout, new_conn->timeout);

	return(new_conn);
}

/* remove a connection from the table */
delete_connection(struct connection *conn){

	if (conn==first_connection)
		first_connection=conn->next;
	if (conn==last_connection)
		last_connection=conn->prev;

	if (conn->prev)
		conn->prev->next=conn->next;
	if (conn->next)
		conn->next->prev=conn->prev;

	free(conn);

}

display_connections() {

	struct connection *conn;
	char *ip_src, *ip_dst, *ip_src_mask, *ip_dst_mask;

	printf("connections table:\n");
	for (conn=first_connection; conn; conn=conn->next) {
		ip_src=strdup(inet_ntoa(conn->src_addr));
		ip_dst=strdup(inet_ntoa(conn->dst_addr));
		ip_src_mask=strdup(inet_ntoa(conn->src_mask));
		ip_dst_mask=strdup(inet_ntoa(conn->dst_mask));
		printf("%s:%s %s:%s %u->%u %u->%u %u %s %lu\n", ip_src,
		       ip_src_mask, ip_dst, ip_dst_mask, conn->proto,
		       conn->src_port_lo, conn->src_port_hi, conn->dst_port_lo,
		       conn->dst_port_hi, conn->interface, conn->timeout);
		/* don't leak memory! */
		free(ip_src);
		free(ip_src_mask);
		free(ip_dst);
		free(ip_dst_mask);

	}

}

expire_connections() {

	struct connection *conn, *next_conn;
	char *ip_src, *ip_dst, *ip_src_mask, *ip_dst_mask;
	time_t the_time;

	conn=first_connection;
	next_timeout=LONG_TIMEOUT;
	the_time=time((time_t)NULL);
	while(conn) {
		next_conn=conn->next;
		if (conn->timeout<0) {
			/* connections with a -ve timeout are for holding
			 * state across packets, ignore them for rules
			 */
		} else if (conn->timeout<=the_time) {
			if (debuglev>0) {
				ip_src=strdup(inet_ntoa(conn->src_addr));
				ip_dst=strdup(inet_ntoa(conn->dst_addr));
				ip_src_mask=strdup(inet_ntoa(conn->src_mask));
				ip_dst_mask=strdup(inet_ntoa(conn->dst_mask));
				debug("expired %s:%s %s:%s %u->%u %u->%u %u %lu\n",
				      ip_src, ip_src_mask, ip_dst, ip_dst_mask,
				      conn->proto, conn->src_port_lo,
				      conn->src_port_hi, conn->dst_port_lo,
				      conn->dst_port_hi, conn->timeout);
				/* don't leak memory! */
				free(ip_src);
				free(ip_src_mask);
				free(ip_dst);
				free(ip_dst_mask);
			}
			delete_rule(conn);
		} else {
			if (debuglev>0) {
				ip_src=strdup(inet_ntoa(conn->src_addr));
				ip_dst=strdup(inet_ntoa(conn->dst_addr));
				ip_src_mask=strdup(inet_ntoa(conn->src_mask));
				ip_dst_mask=strdup(inet_ntoa(conn->dst_mask));
				debug("not expired %s:%s %s:%s %u->%u %u->%u %u %lu\n",
				      ip_src, ip_src_mask, ip_dst, ip_dst_mask,
				      conn->proto, conn->src_port_lo,
				      conn->src_port_hi, conn->dst_port_lo,
				      conn->dst_port_hi, conn->timeout);
				/* don't leak memory! */
				free(ip_src);
				free(ip_src_mask);
				free(ip_dst);
				free(ip_dst_mask);
			}
			next_timeout=MIN(next_timeout, conn->timeout);
		}
		conn=next_conn;
	}
	if (next_timeout<=the_time)
		next_timeout=LONG_TIMEOUT;

}

struct connection *
find_connection(struct in_addr src, struct in_addr src_mask,
		  struct in_addr dst, struct in_addr dst_mask,
		  uint proto, uint src_port_lo, uint src_port_hi,
		  uint dst_port_lo, uint dst_port_hi, char *interface) {

	struct connection *conn, *next_conn;

	conn=first_connection;
	while(conn) {
		next_conn=conn->next;
		if (memcmp(&conn->src_addr, &src, sizeof(src))==0 &&
		    memcmp(&conn->src_mask, &src_mask, sizeof(src))==0 &&
		    memcmp(&conn->dst_addr, &dst, sizeof(dst))==0 &&
		    memcmp(&conn->dst_mask, &dst_mask, sizeof(dst))==0 &&
		    conn->proto==proto &&
		    conn->src_port_lo==src_port_lo &&
		    conn->src_port_hi==src_port_hi &&
		    conn->dst_port_lo==dst_port_lo &&
		    conn->dst_port_hi==dst_port_hi &&
		    (!interface || strcmp(conn->interface, interface)==0)) {
			return(conn);
		}
		conn=next_conn;
	}
	return((struct connection *)NULL);

}

install_spf_filters() {

	static struct ip_fwuser fw;
	int ret, c, line;
	char *v[100], buf[BUFSIZ];
	FILE *rules;

	/* flush input and output chains */
	if ((ret=flush_entries("input"))!=0) {
		fprintf(stderr, "error flushing input rules %s:%d\n",
			__FILE__, __LINE__);
	} 
	if ((ret=flush_entries("output"))!=0) {
		fprintf(stderr, "error flushing output rules %s:%d\n",
			__FILE__, __LINE__);
	} 

	/* install netlink filters on the output chain */
	fw.ipfw.fw_src.s_addr=0L;
	fw.ipfw.fw_smsk.s_addr=0L;
	fw.ipfw.fw_dst.s_addr=0L;
	fw.ipfw.fw_dmsk.s_addr=0L;
	fw.ipfw.fw_proto=IPPROTO_ICMP;
	fw.ipfw.fw_flg=IP_FW_F_NETLINK;
	fw.ipfw.fw_tosand = 0xFF;
	fw.ipfw.fw_tosxor = 0x00;
	fw.ipfw.fw_invflg=IP_FW_INV_VIA;
	/* don't want to see fragments */
	fw.ipfw.fw_flg |= IP_FW_F_FRAG;
	fw.ipfw.fw_invflg |= IP_FW_INV_FRAG;
	fw.ipfw.fw_spts[0]=0;
	fw.ipfw.fw_spts[1]=0xFFFF;
	fw.ipfw.fw_dpts[0]=0;
	fw.ipfw.fw_dpts[1]=0xFFFF;
	fw.ipfw.fw_redirpt=0;
	fw.ipfw.fw_outputsize=64;
	strcpy(fw.ipfw.fw_vianame, "lo");
	strcpy(fw.label, "ACCEPT");
	nsaddrs=1;
	saddrs=&fw.ipfw.fw_src;
	ndaddrs=1;
	daddrs=&fw.ipfw.fw_dst;
	if ((ret=insert_entry("output", &fw, 1, 0, 0))!=0) {
		fprintf(stderr, "error inserting netlink rule %s:%d\n",
			__FILE__, __LINE__);
	} 

	fw.ipfw.fw_proto=IPPROTO_UDP;
	if ((ret=insert_entry("output", &fw, 1, 0, 0))!=0) {
		fprintf(stderr, "error inserting netlink rule %s:%d\n",
			__FILE__, __LINE__);
	} 

#if defined(ALLOW_EST_TCP)
	/* only need to look at TCP output packets for select TCP protos */
	fw.ipfw.fw_src.s_addr=0L;
	fw.ipfw.fw_smsk.s_addr=0L;
	fw.ipfw.fw_dst.s_addr=0L;
	fw.ipfw.fw_dmsk.s_addr=0L;
	fw.ipfw.fw_redirpt=0;
	fw.ipfw.fw_proto=IPPROTO_TCP;
	fw.ipfw.fw_flg=IP_FW_F_NETLINK;
	fw.ipfw.fw_invflg=IP_FW_INV_VIA;
	/* don't want to see fragments */
	fw.ipfw.fw_flg |= IP_FW_F_FRAG;
	fw.ipfw.fw_invflg |= IP_FW_INV_FRAG;
	strcpy(fw.ipfw.fw_vianame, "lo");
	strcpy(fw.label, "ACCEPT");
	nsaddrs=1;
	saddrs=&fw.ipfw.fw_src;
	ndaddrs=1;
	daddrs=&fw.ipfw.fw_dst;

	/* FTP */
	fw.ipfw.fw_spts[0]=21;
	fw.ipfw.fw_spts[1]=21;
	fw.ipfw.fw_dpts[0]=0;
	fw.ipfw.fw_dpts[1]=0xFFFF;
	fw.ipfw.fw_outputsize=128;
	if ((ret=insert_entry("output", &fw, 1, 0, 0))!=0) {
		fprintf(stderr, "error inserting netlink rule %s:%d\n",
			__FILE__, __LINE__);
	} 
	/* FTP */
	fw.ipfw.fw_spts[0]=0;
	fw.ipfw.fw_spts[1]=0xFFFF;
	fw.ipfw.fw_dpts[0]=21;
	fw.ipfw.fw_dpts[1]=21;
	fw.ipfw.fw_outputsize=128;
	if ((ret=insert_entry("output", &fw, 1, 0, 0))!=0) {
		fprintf(stderr, "error inserting netlink rule %s:%d\n",
			__FILE__, __LINE__);
	} 
	/* RealAudio */
	fw.ipfw.fw_spts[0]=0;
	fw.ipfw.fw_spts[1]=0xFFFF;
	fw.ipfw.fw_dpts[0]=554;
	fw.ipfw.fw_dpts[1]=554;
	fw.ipfw.fw_outputsize=128;
	if ((ret=insert_entry("output", &fw, 1, 0, 0))!=0) {
		fprintf(stderr, "error inserting netlink rule %s:%d\n",
			__FILE__, __LINE__);
	} 
	fw.ipfw.fw_spts[0]=0;
	fw.ipfw.fw_spts[1]=0xFFFF;
	fw.ipfw.fw_dpts[0]=7070;
	fw.ipfw.fw_dpts[1]=7070;
	fw.ipfw.fw_outputsize=128;
	if ((ret=insert_entry("output", &fw, 1, 0, 0))!=0) {
		fprintf(stderr, "error inserting netlink rule %s:%d\n",
			__FILE__, __LINE__);
	} 
#else
	/* look at all output TCP packets */
	fw.ipfw.fw_proto=IPPROTO_TCP;
	if ((ret=insert_entry("output", &fw, 1, 0, 0))!=0) {
		fprintf(stderr, "error inserting netlink rule %s:%d\n",
			__FILE__, __LINE__);
	} 
#endif

#if defined(ALLOW_EST_TCP) && !defined(ALLOW_EST_TCP_NORULE)
	/* install a rule for established TCP on the input chain */
	fw.ipfw.fw_src.s_addr=0L;
	fw.ipfw.fw_smsk.s_addr=0L;
	fw.ipfw.fw_dst.s_addr=0L;
	fw.ipfw.fw_dmsk.s_addr=0L;
	fw.ipfw.fw_proto=IPPROTO_TCP;
	fw.ipfw.fw_flg=IP_FW_F_TCPSYN;
	fw.ipfw.fw_tosand = 0xFF;
	fw.ipfw.fw_tosxor = 0x00;
	fw.ipfw.fw_invflg=IP_FW_INV_SYN;
	fw.ipfw.fw_spts[0]=0;
	fw.ipfw.fw_spts[1]=0xFFFF;
	fw.ipfw.fw_dpts[0]=0;
	fw.ipfw.fw_dpts[1]=0xFFFF;
	fw.ipfw.fw_redirpt=0;
	fw.ipfw.fw_outputsize=0;
	strcpy(fw.ipfw.fw_vianame, SPF_VIANAME);
	strcpy(fw.label, "ACCEPT");
	nsaddrs=1;
	saddrs=&fw.ipfw.fw_src;
	ndaddrs=1;
	daddrs=&fw.ipfw.fw_dst;
	if ((ret=insert_entry("input", &fw, 1, 0, 0))!=0) {
		fprintf(stderr, "error inserting netlink rule %s:%d\n",
			__FILE__, __LINE__);
	} 
#endif

	/*
	 * install user defined rules in their own chain
	 */

	delete_chain("statinpt", 0);	/* try to remove the chain first */
	if ((ret=new_chain("statinpt"))==-1) {	/* create the new chain */
		perror("Creating statinpt chain");
		cleanup_exit(SIGTERM);
	}
	
	/* and link it as the last thing to do in the input chain */
	fw.ipfw.fw_src.s_addr=0L;
	fw.ipfw.fw_smsk.s_addr=0L;
	fw.ipfw.fw_dst.s_addr=0L;
	fw.ipfw.fw_dmsk.s_addr=0L;
	fw.ipfw.fw_proto=0;
	fw.ipfw.fw_flg=0;
	fw.ipfw.fw_tosand = 0xFF;
	fw.ipfw.fw_tosxor = 0x00;
	fw.ipfw.fw_invflg=0;
	fw.ipfw.fw_spts[0]=0;
	fw.ipfw.fw_spts[1]=0xFFFF;
	fw.ipfw.fw_dpts[0]=0;
	fw.ipfw.fw_dpts[1]=0xFFFF;
	fw.ipfw.fw_redirpt=0;
	fw.ipfw.fw_outputsize=0;
	strcpy(fw.ipfw.fw_vianame, SPF_VIANAME);
	strcpy(fw.label, "statinpt");
	nsaddrs=1;
	saddrs=&fw.ipfw.fw_src;
	ndaddrs=1;
	daddrs=&fw.ipfw.fw_dst;
	if ((ret=insert_entry("input", &fw, START_DYN_RULES, 0, 0))!=0) {
		fprintf(stderr, "error inserting netlink rule %s:%d\n",
			__FILE__, __LINE__);
	} 
	
	if ((rules=fopen("/etc/spf-rules", "r"))==NULL) {
		if (errno==ENOENT)
			return(0);
		else {
			perror("Opening rules file");
			cleanup_exit(SIGTERM);
		}
	}
	/* read a line at a time, vectorizing args into an array */
	c=0;
	v[c]="ipchains";
	line=0;
	while (!feof(rules)){
		
		char *s, *b;

		line++;
		if (fgets(buf, BUFSIZ, rules)==NULL)
			break;
		/* strip off comments */
		if (s=strchr(buf, '#'))
			*s='\0';
		if (s=strchr(buf, '\n'))
			*s='\0';	/* get rid of \n at the end */
		/* skip the line if it's blank */
		if (buf[0]=='\0')
			continue;
		b=buf;
		while (1) {
			v[++c]=b;
			if ((s=strchr(b, ' '))==NULL)
				break;
			while (*s==' ')
				*s++='\0';
			b=s;
			/* change any input chain specifiers to statinpt */
			if (strcmp(v[c], "input")==0)
				if (strcmp(v[c-1], "-A")==0 ||
				    strcmp(v[c-1], "-D")==0 ||
				    strcmp(v[c-1], "-R")==0 ||
				    strcmp(v[c-1], "-I")==0 ||
				    strcmp(v[c-1], "-P")==0)
					v[c]="statinpt";
		}
		/* check to see if we got anything at all */
		if (c==0)
			continue;	/* nope, next line please */
#if 0
		/* remove the \n from the last arg */
		v[c++][strlen(v[c])-1]='\0';
#else
		c++;
#endif
		/* run the installer */
		if (ipchains(c, v)!=0) {
			fprintf(stderr, "error installing rule on line %d\n",
				line);
			cleanup_exit(SIGTERM);
		}
		c=0;
	}
	fclose(rules);
	
	/* install default deny filter on the input chain */
	fw.ipfw.fw_src.s_addr=0L;
	fw.ipfw.fw_smsk.s_addr=0L;
	fw.ipfw.fw_dst.s_addr=0L;
	fw.ipfw.fw_dmsk.s_addr=0L;
	fw.ipfw.fw_proto=0;
	fw.ipfw.fw_flg=IP_FW_F_PRN;
	fw.ipfw.fw_tosand = 0xFF;
	fw.ipfw.fw_tosxor = 0x00;
	fw.ipfw.fw_invflg=0;
	fw.ipfw.fw_spts[0]=0;
	fw.ipfw.fw_spts[1]=0xFFFF;
	fw.ipfw.fw_dpts[0]=0;
	fw.ipfw.fw_dpts[1]=0xFFFF;
	fw.ipfw.fw_redirpt=0;
	fw.ipfw.fw_outputsize=0;
	strcpy(fw.ipfw.fw_vianame, SPF_VIANAME);
	strcpy(fw.label, DEFAULT_ACTION);
	nsaddrs=1;
	saddrs=&fw.ipfw.fw_src;
	ndaddrs=1;
	daddrs=&fw.ipfw.fw_dst;
	if ((ret=append_entry("input", &fw, 0, 0))!=0) {
		fprintf(stderr, "error inserting netlink rule %s:%d\n",
			__FILE__, __LINE__);
	} 

}

/* accumulate time since last call to "timer" */
void
set_timer(struct timeval *timer) {

	/* to remember the time when last called */
	static struct timeval last;
	struct timeval now;
	struct timezone tz;
	long sec, usec;

	/* What time is it? */
	gettimeofday(&now, &tz);
	
	/* see if this was a call to reset the "last" timer */
	if (!timer) {
		last.tv_sec=now.tv_sec;
		last.tv_usec=now.tv_usec;
		return;
	}
	/* subtract last time */
	sec=(now.tv_sec-last.tv_sec);
	if ((usec=(now.tv_usec-last.tv_usec))<1) {
		sec--;
		usec+=1000000;
	}
	timer->tv_sec+=sec;
	if ((timer->tv_usec+=usec)>1000000) {
		timer->tv_sec++;
		timer->tv_usec-=1000000;
	}
	last.tv_sec=now.tv_sec;
	last.tv_usec=now.tv_usec;
		
}

/* convert the value of a timer to a char * */
char *
timer2str(struct timeval *timer) {

	static char s[40];

	sprintf(s, "%d.%07ld", timer->tv_sec, timer->tv_usec);
	return(s);

}
